/*  aqotdd - Afghan's Quote of the day server - An RFC865 compliant server.
    Copyright (C) 2000 Afghan Hound Software

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/

/* qotd.c - sets up data structures (with dir.c) and actually finds quotes
   from fortune data files */

#include "main.h"

/* Only need the rot13'd flag from strfile */
#define QOTD_ROTATED               0x4                 
#define QOTD_HEADER_VERSION        1
#define QOTD_DELIM_CHAR            stuff[0]      

/* Fortune file struct - from man strfile */
typedef struct {                               
  unsigned long   str_version;            /* Always == 1? */
  unsigned long   str_numstr;             
  unsigned long   str_longlen;            /* Don't use these - but we only(!) */
  unsigned long   str_shortlen;           /* waste 8 bytes per fortune file...*/
  unsigned long   str_flags;              /* rot13 (QOTD_ROTATED) == offensive */
  unsigned char   stuff[4];               /* first char is the deliminator    */
} qotd_fileheader;

/* QOTD Variables */
char            **qotd_dirs;            /* Directories with qotds    */
char            **qotd_files;           /* Valid qotd files          */
long int        **qotd_offsets;         /* Quote offsets             */
qotd_fileheader **qotd_headers;         /* File headers              */
int               qotd_num_files   = 0; /* Number of files found     */
int               qotd_num_dirs    = 0; /* Number of sub directories */
long int          qotd_quote_total = 0; /* Total number of quotes    */

/* Local Functions */
void qotd_free_dir_struct();
void qotd_free_file_struct();
int  qotd_check_fileheader(char *cwd,char *filename);
void qotd_load_fileheader(char *cwd,char *filename,qotd_fileheader *header);
void qotd_file_add(char *cwd,char *filename);
void qotd_list_files(char *cwd);
void qotd_find_subdirs(char *current);


void qotd_free_dir_struct()
{
  int i;
  if (qotd_num_dirs)
    {
      for(i=(qotd_num_dirs-1);i>=0;i--)
	{
	  if (qotd_dirs[i]!=NULL){free(qotd_dirs[i]);}
	}
    }
  free(qotd_dirs);
}

void qotd_free_file_struct()
{
  int i;
  if (qotd_num_files)
    {
      for(i=(qotd_num_files-1);i>=0;i--)
	{
	  if (qotd_files[i]!=NULL){free(qotd_files[i]);}
	  if (qotd_headers[i]!=NULL){free(qotd_headers[i]);}
	  if (qotd_offsets[i]!=NULL){free(qotd_offsets[i]);}
	}
    }
  free(qotd_offsets);
  free(qotd_files);
  free(qotd_headers);
}

void qotd_load_fileheader(char *cwd, char *filename,qotd_fileheader *header)
{
  char         *fullname;
  FILE         *fptr;
  int           ret;
  unsigned long num,i;

  fullname=(char *)xmalloc((strlen(filename)+1+4+1+strlen(cwd)+1)*sizeof(char));
  sprintf(fullname,"%s/%s.dat",cwd,filename);

  if ((fptr = fopen(fullname,"r"))==NULL)
    {
      my_message(MSG_ERRFILEF,"File gone away.",fullname);
    }
  ret = fread(header,sizeof(qotd_fileheader),1,fptr);
  if (!(ret))
    {
      my_message(MSG_ERRFILEF,"File gone away",fullname);
    }
  num = ntohl(header->str_numstr);
  qotd_offsets[qotd_num_files]=(long int *)xmalloc(num*sizeof(long int));
  for (i=0;i<num;i++)
    {
      ret = fread(&qotd_offsets[qotd_num_files][i],sizeof(long),1,fptr);
      qotd_offsets[qotd_num_files][i] = ntohl(qotd_offsets[qotd_num_files][i]);
      if (!(ret))
	{
	  my_message(MSG_ERRFILEF,"Error in dat file",fullname);
	}
    }
  fclose(fptr);
  free(fullname);
}

int qotd_check_fileheader(char *cwd, char *filename)
{
  char *fullname;
  FILE *fptr;
  int ret;
  qotd_fileheader *temp1;
  unsigned long u1,u2,u3,u4,u5;
  char delim;
  
  fullname=(char *)xmalloc((strlen(filename)+1+4+1+strlen(cwd)+1)*sizeof(char));
  sprintf(fullname,"%s/%s.dat",cwd,filename);
  if ((fptr = fopen(fullname,"r"))==NULL){return(0);}
  temp1=(qotd_fileheader *)xmalloc(sizeof(qotd_fileheader));
  ret = fread(temp1,sizeof(qotd_fileheader),1,fptr);
  
  u1 = ntohl(temp1->str_version);
  u2 = ntohl(temp1->str_numstr);
  u3 = ntohl(temp1->str_longlen);
  u4 = ntohl(temp1->str_shortlen);
  u5 = ntohl(temp1->str_flags);
  delim = temp1->stuff[0];
  ret =( (ret ==1 ) &&(u1  == QOTD_HEADER_VERSION ) &&
	 (u2  > 0 ) && (u2 <65535) &&
	 (u3  > 0 ) && (u3 <65535) &&
	 (u4  > 0 ) && (u4 <65535) &&
	 (u5 <8)       );
  
  free (temp1);
  fclose(fptr);
  free(fullname);
  return(ret);
}


void qotd_file_add(char *cwd,char *filename)
{  
  if (!qotd_check_fileheader(cwd,filename))
    {
      return;
    }

  if (!(qotd_num_files))
    {
      qotd_files=(char **)xmalloc(sizeof(char *));
      qotd_offsets=(long int **)xmalloc(sizeof(long int *));
      qotd_headers=(qotd_fileheader **)xmalloc(sizeof(qotd_fileheader *));
    }
  else
    {
      qotd_files   = (char **)xrealloc(qotd_files,(qotd_num_files+1)*sizeof(char *));
      qotd_offsets = (long int **)xrealloc(qotd_offsets,(qotd_num_files+1)*sizeof(long int *));
      qotd_headers = (qotd_fileheader **)xrealloc(qotd_headers,(qotd_num_files+1)*sizeof(qotd_fileheader *));
    }
  qotd_files[qotd_num_files]=(char *)xmalloc((strlen(filename)+1+strlen(cwd)+1+1)*sizeof(char));
  qotd_headers[qotd_num_files]=(qotd_fileheader *)xmalloc(sizeof(qotd_fileheader));
  sprintf(qotd_files[qotd_num_files],"%s/%s",cwd,filename);
  qotd_load_fileheader(cwd,filename,qotd_headers[qotd_num_files]);
  qotd_num_files++;
}

void qotd_list_files(char *cwd)
{
  DIR *dir_ptr1;
  DIR *dir_ptr2;
  char *filename;
  struct dirent *next_entry;
  struct dirent *dat_poss;
  int flag1,flag2;

  if ( ((dir_ptr1=opendir(cwd))==NULL) || ((dir_ptr2=opendir(cwd))==NULL))
    {
      my_message(MSG_ERRFILEC,"Opening Extra DIR, ignoring...",strerror(errno));
      return;
    } 
  while ((next_entry = readdir (dir_ptr1)))
    {
      if ((strcmp(next_entry->d_name,".."))&&(strcmp(next_entry->d_name,"."))&&((strstr(next_entry->d_name,".dat"))==NULL))
	{
	  filename=(char *)xmalloc((strlen(next_entry->d_name)+1+4)*sizeof(char));
	  sprintf(filename,"%s.dat",next_entry->d_name);
	  flag1 = 0;
	  flag2 = 0;
	  rewinddir(dir_ptr2);
	  while( (!flag1) && (!flag2) )
	    {
	      flag2 = ((dat_poss=readdir(dir_ptr2))==NULL);
	      if(!flag2){flag1 = (!(strcmp(dat_poss->d_name,filename)));}
	    }
	  if (flag1){qotd_file_add(cwd,next_entry->d_name);}
	  free(filename);
	}
    }
  closedir(dir_ptr1);
  closedir(dir_ptr2);
}


void qotd_find_subdirs(char *current)
{
  DIR           *current_dir;
  DIR           *next_dir;
  struct dirent *next_entry;
  struct stat   *stat_buf;
  char          *nextfilename;
  int            ret;

  stat_buf=(struct stat*)xmalloc(sizeof(struct stat));
  if((current_dir=opendir(current))==NULL)
    {
      my_message(MSG_ERRFILEC,"Opening Extra DIR, ignoring...",strerror(errno));
      free(stat_buf);
      return;
    } 
  while ((next_entry = readdir (current_dir)))
    {
      if ((strcmp(next_entry->d_name,".."))&&(strcmp(next_entry->d_name,".")))
	{
	  nextfilename=(char *)xmalloc((strlen(current)+3+strlen(next_entry->d_name))*sizeof(char));
	  sprintf(nextfilename,"%s/%s",current,next_entry->d_name);
	  ret = stat(nextfilename, stat_buf); 
	  if ((!ret)&&(stat_buf!=NULL)&&(S_ISDIR(stat_buf->st_mode)))
	    { 
	      if((next_dir=opendir(nextfilename))==NULL)
		{
		  my_message(MSG_ERRFILEC,"Opening Next DIR",strerror(errno));
		}
	      else
		{
		  qotd_dirs=(char **)xrealloc(qotd_dirs,(qotd_num_dirs+1)*sizeof(char *));
		  qotd_dirs[qotd_num_dirs]=(char *)xmalloc((strlen(nextfilename)+1)*sizeof(char));
		  strcpy(qotd_dirs[qotd_num_dirs],nextfilename);
		  qotd_num_dirs++;
		  closedir(next_dir);
		  qotd_find_subdirs(nextfilename);
		}
	    }
	  free(nextfilename);
	}
    }
  closedir(current_dir);
  free(stat_buf);
}


void qotd_init()
{
  int i;
  DIR *qotd_dir_ptr;
  
  my_message(MSG_INITSTART,"QOTD DB init starting.","");
  srand(time(NULL));
  qotd_free_dir_struct();
  qotd_num_dirs = 0;
  if((qotd_dir_ptr=opendir(qotd_dir))==NULL)
    {
      my_message(MSG_ERRCONT,"Opening Main DIR",strerror(errno));
      qotd_free_file_struct();
      qotd_num_files = 0;
      qotd_quote_total = 0;
      my_message(MSG_ERRCONT,"No fortunes found!","");
      my_message(MSG_INITFINISH,"QOTD DB init done.","");
      return;
    }
  qotd_dirs   =(char **)xmalloc(sizeof(char *));
  qotd_dirs[0]=(char *) xmalloc((strlen(qotd_dir)+1)*sizeof(char));
  strcpy(qotd_dirs[0],qotd_dir);
  qotd_num_dirs++;
  closedir(qotd_dir_ptr);
  qotd_find_subdirs(qotd_dir);  
  qotd_free_file_struct();
  qotd_num_files = 0;
  for (i=0;i<qotd_num_dirs;i++)
    {
      qotd_list_files(qotd_dirs[i]);
    }
  qotd_quote_total = 0;
  for (i=0;i<qotd_num_files;i++)
    {
      qotd_quote_total += ntohl(qotd_headers[i]->str_numstr);
    }
  if (qotd_quote_total == 0){
    my_message(MSG_ERRCONT,"No fortunes found!","");
  }
  my_message(MSG_INITFINISH,"QOTD DB init done.","");
}

void unrot(char *buffer)
{
  unsigned int i;
  for (i=0;i<strlen(buffer);i++)
    {
      if (isupper((int)buffer[i]))
	{
	  buffer[i] = 'A' + (buffer[i] - 'A' + 13) %26;
	}
      if (islower((int)buffer[i]))
	{
	  buffer[i] = 'a' + (buffer[i] - 'a' + 13) %26;
	}
    }
}

void get_file_offset(int fnum,int *file_num,int *offset)
{
  int t,f;
  f = 0;

  if (fnum == 0)
    {
      file_num[0] = 0;
      offset[0] = 0;
      return;
    }

  while((fnum>0)&&(f<qotd_num_files))
    {
      fnum -= ntohl(qotd_headers[f]->str_numstr);
      f++;
    }
  f--;
  t = fnum+ntohl(qotd_headers[f]->str_numstr);
  file_num[0] = f;offset[0] = t;
}

char *qotd_getquote(char *buffer)
{
  FILE     *fort;
  int      i,j,limit,grow,file_num,offset,fortune_num,flag1;
  long int flags;
  char      nextchar,lastchar;

  buffer = NULL;
  grow=250; limit=250; /* Experimentally determined */
  i = 0;flag1 = 0;
  
  if (!qotd_quote_total)
    {
      buffer = (char *)xmalloc(300*sizeof(char));
      strcpy(buffer,"Afghan's QOTD server can find no fortunes. :(\n");
      return buffer;
    }

  fortune_num = (int)((1.0*rand()*qotd_quote_total/(RAND_MAX+1.0)));

  

  get_file_offset(fortune_num,&file_num,&offset);
  flags = ntohl(qotd_headers[file_num]->str_flags);
  if ((flags & QOTD_ROTATED)&&(!(use_off)))
    {
      buffer = qotd_getquote(buffer);
      return buffer;
    }

  if ((fort = fopen(qotd_files[file_num],"r"))==NULL)
    {
      my_message(MSG_ERRFILEC,"Error opening fortune file. Re-initialising\n","");
      qotd_init();
      buffer = qotd_getquote(buffer);
      return buffer;
    }
 
  if ((fseek(fort,qotd_offsets[file_num][offset],SEEK_SET)) < 0)
    {
      my_message(MSG_ERRFILEC,"Error seeking in fortune file. Re-initialising\n","");
      fclose(fort);
      qotd_init();
      buffer = qotd_getquote(buffer);
      return buffer;
    } 
  
  buffer=xmalloc(limit*sizeof(char));

  nextchar = fgetc(fort);
  lastchar = ' ';
  while( !(
	   (lastchar==qotd_headers[file_num]->QOTD_DELIM_CHAR) && 
	   ((nextchar=='\n')||(nextchar==EOF)))    
	 &&
	 (nextchar!=EOF) 
	 )
    {
      buffer[i]=nextchar;i++;
      if (nextchar=='\t'){i+=(TABSTOP-1);}
      if (i>=(limit-(TABSTOP))) /*i could have moved too far */
	{
	  realloc1++;
	  realloc2+=flag1;
	  flag1=1;
	  buffer=xrealloc(buffer,limit+grow);
	  limit +=grow;
	}
      if(nextchar=='\t')
        {
            for(j=(i-TABSTOP);j<i;j++)
               {
                  buffer[j]=' ';
               }
            nextchar   =' ';
        }
      if (((i>=512)&&(use_rfc))||(i>=abs_max)) /* Abs max */
	{
	  /*too long*/
	  fclose(fort);
	  free(buffer);
	  buffer=qotd_getquote(buffer);
	  return buffer;
	}
      lastchar=nextchar;
      nextchar = fgetc(fort);
      if (use_rfc)
	{
	  /* RFC says - printing/space/CR/LF only */
	  while (
		 (!isprint((int)nextchar)) &&
		 (!(nextchar=='\n'))   &&
		 (!(nextchar==' '))    && 
		 (!(nextchar==EOF))    &&
		 (!(nextchar=='\t'))    /* Let tabs through - we expand them*/
		 )
	    {
	      nextchar=fgetc(fort);
	    }
	}
    }
  fclose(fort);
  if ((i<abs_min)||(i<2))
    {
      if (buffer!=NULL)
	{
	  free(buffer);
	}
      buffer=qotd_getquote(buffer);
      return buffer;
    }
  buffer[(i-2)] = '\n';
  buffer[(i-1)] = 0;
  buffer[i]     = 0;
  if (flags & QOTD_ROTATED)
    {
      unrot(buffer);
    }
  return buffer;
}
